/*
 * This script moves + rotates the current selection so that
 * the track end selected by user connects to another track. The
 * original track is removed.
 * Attention! This script is highly re-used by DupMove, don't
 * change it without checking DupMove!
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-03-30
 */

include("../QCTrack.js");

function DupParallel(guiAction) {
    EAction.call(this, guiAction);
    if (!isNull(guiAction)) {
      this.setUiOptions([DupParallel.includeBasePath + "/DupParallel.ui"]);
    }
}

DupParallel.prototype = new EAction();
DupParallel.includeBasePath = includeBasePath;

DupParallel.State = {
    SettingLineSource : 0
};

DupParallel.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    this.setState(DupParallel.State.SettingLineSource);
};

DupParallel.prototype.setState = function(state) {
    EAction.prototype.setState.call(this, state);

    this.getDocumentInterface().setClickMode(RAction.PickEntity);
    this.setCrosshairCursor();

    var appWin = RMainWindowQt.getMainWindow();
    var info;
    switch (this.state) {
    case DupParallel.State.SettingLineSource:
      this.lineSource = undefined;
      info = qsTr("Select track to duplicate");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trCancel);
      break;
    }
};

DupParallel.prototype.escapeEvent = function() {
    switch (this.state) {
    case DupParallel.State.SettingLineSource:
      EAction.prototype.escapeEvent.call(this);
      break;
    }
};

DupParallel.prototype.pickEntity = function(event, preview) {
    var di = this.getDocumentInterface();
    var document = this.getDocument();
    var entityId = event.getEntityId();
    var entity = document.queryEntity(entityId);
    var pos = event.getModelPosition();
    var op;

    switch (this.state) {
    case DupParallel.State.SettingLineSource:
      if (isNull(entity)) {
        this.lineSource = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineSource = getClosestTrackStraightMid(document, entity, pos, false);
      this.sourceBlock = undefined;
      this.lineSide = undefined;
      if (this.lineSource == undefined) {
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineSide = this.lineSource.getSideOfPoint(pos);
      this.sourceBlock = entity;
      if (preview) {
        this.updatePreview();
      }
      else {
        op = this.getOperation(false);
        if (op == undefined) {
          break;
        }
        di.applyOperation(op);
      }
      break;
    }
};

DupParallel.prototype.getOperation = function(preview) {
    var document = this.getDocument();
    
    if (isNull(document) || !isLineEntity(this.lineSource) || this.sourceBlock == undefined || this.lineSide == undefined) {
      return undefined;
    }

    var op = new RAddObjectsOperation();
    var blockReferenceData = new RBlockReferenceData();
    blockReferenceData.setReferencedBlockId(this.sourceBlock.getReferencedBlockId());
    var dest = this.sourceBlock.getData().getPosition();
    var offset = this.getOffset();
    blockReferenceData.setPosition(dest);
    blockReferenceData.setScaleFactors(new RVector(1.0, 1.0));
    blockReferenceData.setRotation(this.sourceBlock.getRotation());
    blockReferenceData.move(offset);
    var blockRef = new RBlockReferenceEntity(document, blockReferenceData);
    op.addObject(blockRef);
    return op;
};

DupParallel.prototype.getAuxPreview = function() {
    var ret = undefined;
    if (isLineEntity(this.lineSource) && this.sourceBlock != undefined) {
      ret = new Array();
      var mid = getPointOnLine(this.lineSource, 0.5);
      var angle = this.sourceBlock.getRotation();
      ret.push(new RArc(mid, this.distance, angle, angle + Math.PI, this.lineSide == RS.RightHand));
      var offset = this.getOffset();
      var line = new RLine(this.lineSource.getStartPoint(), this.lineSource.getEndPoint());
      line.move(offset);
      ret.push(line);
    }
    return ret;
};

DupParallel.prototype.getOffset = function() { //call only with valid this.sourceBlock and this.lineSide!
    var line = new RLine(new RVector(0, 0), new RVector(0, this.distance));
    line.rotate(this.sourceBlock.getRotation());
    if (this.lineSide == RS.RightHand) {
      line.rotate(Math.PI);
    }
    return line.getEndPoint();
};

DupParallel.prototype.slotDistanceChanged = function(value) {
    this.distance = value;
    this.updatePreview(true);
};

DupParallel.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.DupParallel, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/DupParallel.js");
    action.setIcon(basePath + "/DupParallel.svg");
    action.setDefaultShortcut(new QKeySequence("q,p"));
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupModification);
    action.setSortOrder(QCTrack.MenuOrder.DupParallel);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

