/*
 * This script ensures the object order of track layer objects.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-01-26
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");
include("../../WidgetFactory.js");

function EnsureOrder(guiAction) {
    EAction.call(this, guiAction);
}

EnsureOrder.prototype = new EAction();
EnsureOrder.includeBasePath = includeBasePath;

EnsureOrder.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();
    var storage = EAction.getStorage();

    var nmbrs = moveObjectsToBack(document, di, storage, "TR_NMBRS");
    var plan = moveObjectsToBack(document, di, storage, "TR_PLAN");
    var swells = moveObjectsToBack(document, di, storage, "TR_SWELLS");
    var beds = moveObjectsToBack(document, di, storage, "TR_BEDS");
    var ends = moveObjectsToBack(document, di, storage, "TR_ENDS");
    var mids = moveObjectsToBack(document, di, storage, "TR_MIDS");
    var cost = moveObjectsToBack(document, di, storage, "TR_COST");

    //--cleanup
    EAction.handleUserMessage(qsTr("object ordering done with objects: ")
      + nmbrs + qsTr(" on TR_NMBRS, ") + plan + qsTr(" on TR_PLAN, ") + swells + qsTr(" on TR_SWELLS, ") 
      + beds + qsTr(" on TR_BEDS, ") + ends + qsTr(" on TR_ENDS, ") + mids + qsTr(" on TR_MIDS, ") + cost + qsTr(" on TR_COST"));
    this.terminate();
};

EnsureOrder.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.EnsureOrder, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/EnsureOrder.js");
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupCreation);
    action.setSortOrder(QCTrack.MenuOrder.EnsureOrder);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

