/*
 * This script creates a circle for all "unconnected" track ends.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-01-25
 */

include("../QCTrack.js");

function MarkOpen(guiAction) {
    EAction.call(this, guiAction);
}

MarkOpen.prototype = new EAction();
MarkOpen.includeBasePath = includeBasePath;

MarkOpen.prototype.beginEvent = function() {
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();

    var dialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/MarkOpen",
                   "MarkOpenDialog.ui");
    WidgetFactory.restoreState(dialog);
    var widgets = getWidgets(dialog);
    widgets["buttonBox"].button(QDialogButtonBox.RestoreDefaults)["clicked()"].connect(this, function() { WidgetFactory.restoreState(dialog, null, null, true); });
    widgets["buttonBox"].button(QDialogButtonBox.Ok).setFocus();
    if (!dialog.exec()) { //the user aborted
        dialog.destroy();
        this.terminate();
        return;
    }
    WidgetFactory.saveState(dialog);
    this.maxDist = widgets["distanceRestrict"].getValue();
    dialog.destroy();
    if (this.maxDist < 0) {
      EAction.handleUserMessage(qsTr("distance must not be negative"));
      (new QMessageBox(null, qsTr("Mark Open Track Ends"), qsTr("distance must not be negative"),
        new QMessageBox.StandardButtons(QMessageBox.Ok), null, Qt.Dialog)).exec();
      this.terminate();
      return;
    }

    removeLayer(document, di, "TR_OPEN");
    checkLayerExistence(document, di, "TR_OPEN", "red");
    this.addSelectionIds = new Array();
    var allIds = document.queryAllBlockReferences();
    var allEnds = new Array(); //entries are all TR_END-lines inside a block
    for (var i = 0; i < allIds.length; i++) { //first run: copy all TR_END-lines in blocks
      var id = allIds[i];
      var block = document.queryEntity(id);
      var data = block.getData();
      var subIds = document.queryBlockEntities(data.getReferencedBlockId());
      var lines = new Array();
      for (var j = 0; j < subIds.length; j++) {
        var line = data.queryEntity(subIds[j]);
        if (isLineEntity(line) && line.getLayerName() == "TR_ENDS") {
          allEnds.push(line);
        }
      }
    }
    var cmp = 0;
    while (cmp < allEnds.length) { //second run: check "duplicate" lines
      var removeCmp = false;
      var to = cmp + 1;
      while (to < allEnds.length) {
        if (allEnds[cmp].getStartPoint().getDistanceTo(allEnds[to].getEndPoint()) <= this.maxDist
            && allEnds[cmp].getEndPoint().getDistanceTo(allEnds[to].getStartPoint()) <= this.maxDist) {
          allEnds.splice(to, 1);
          removeCmp = true;
        }
        else {
          to++;
        }
      }
      if (removeCmp) {
        allEnds.splice(cmp, 1);
      }
      else {
        cmp++;
      }
    }
    di.setCurrentColor(new RColor(RColor.ByLayer));
    di.setCurrentLayer("TR_OPEN");
    var operation = new RAddObjectsOperation();
    for (var i = 0; i < allEnds.length; i++) { //third run: create a circle for all remaining lines
      var start = allEnds[i].getStartPoint();
      var end = allEnds[i].getEndPoint();
      operation.addObject(new RCircleEntity(document, new RCircleData(
        new RVector(start.getX() * 0.5 + end.getX() * 0.5, start.getY() * 0.5 + end.getY() * 0.5),
        start.getDistanceTo(end))));
    }
    di.applyOperation(operation);
    EAction.handleUserMessage(qsTr("open track end count: ") + allEnds.length);

    di.clearPreview();
    di.repaintViews();
    this.terminate();
};

MarkOpen.prototype.recursiveSelect = function(allTracks, cmp) {
    for (var i = 0; i < allTracks.length; i++) {
      if (i == cmp || allTracks[i][2]) {
        continue;
      }
      this.selectIfConnected(allTracks, cmp, i);
    }
};

MarkOpen.prototype.selectIfConnected = function(allTracks, cmp, to) {
  if (this.sameLayer && allTracks[cmp][0].getLayerName() != allTracks[to][0].getLayerName()) {
    return;
  }
  var cmpLines = allTracks[cmp][1];
  var toLines = allTracks[to][1];
  for (var i = 0; i < cmpLines.length; i++) {
    var cmpLine = cmpLines[i];
    for (var j = 0; j < toLines.length; j++) {
      var line = toLines[j];
      if (line.getStartPoint().getDistanceTo(cmpLine.getEndPoint()) <= this.maxDist
          && line.getEndPoint().getDistanceTo(cmpLine.getStartPoint()) <= this.maxDist) {
        this.addSelectionIds.push(allTracks[to][0].getId());
        allTracks[to][2] = true;
        this.recursiveSelect(allTracks, to);
        return;
      }
    }
  }
};

MarkOpen.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.MarkOpen, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/MarkOpen.js");
    action.setIcon(basePath + "/MarkOpen.svg");
    action.setDefaultShortcut(new QKeySequence("q,k"));
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupModification);
    action.setSortOrder(QCTrack.MenuOrder.MarkOpen);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
    var appWin = EAction.getMainWindow();
    appWin.addSelectionListener(action);
};

