/*
 * This script order the objects on all specified layers.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-03-23
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");
include("../../WidgetFactory.js");

function ReorderObjects(guiAction) {
    EAction.call(this, guiAction);
}

ReorderObjects.prototype = new EAction();
ReorderObjects.includeBasePath = includeBasePath;

ReorderObjects.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();
    var storage = EAction.getStorage();

    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();

    //let the user enter the desired parameters
    var dialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/ReorderObjects",
                   "ReorderObjectsDialog.ui");
    this.widgets = getWidgets(dialog);
    this.widgets["listAvail"].itemSelectionChanged.connect(this, "checkButtonEnable");
    this.widgets["listAvail"].itemDoubleClicked.connect(this, "doSort");
    this.widgets["listSort"].itemSelectionChanged.connect(this, "checkButtonEnable");
    this.widgets["listSort"].itemDoubleClicked.connect(this, "noSort");
    this.widgets["goTop"]["clicked()"].connect(this, "goTop");
    this.widgets["goUp"]["clicked()"].connect(this, "goUp");
    this.widgets["goDown"]["clicked()"].connect(this, "goDown");
    this.widgets["goBottom"]["clicked()"].connect(this, "goBottom");
    this.widgets["invertOrder"]["clicked()"].connect(this, "invertOrder");
    this.widgets["noSort"]["clicked()"].connect(this, "noSort");
    this.widgets["doSort"]["clicked()"].connect(this, "doSort");

    var layerIds = document.queryAllLayers();
    this.listAvail = new Array();
    this.listSort = new Array();
    for (var i = 0; i < layerIds.length; i++) {
      var layer = document.queryLayer(layerIds[i]);
      if (layer.isFrozen() || layer.isLocked()) {
        this.listAvail.push(layer.getName());
      }
      else {
        this.listSort.push(layer.getName());
      }
    }
    this.listAvail.sort();
    this.listSort.sort();
    this.rebuildLists();

    if (!dialog.exec()) { //the user aborted
        dialog.destroy();
        this.terminate();
        return;
    }

    var cnt = 0;
    for (var i = 0; i < this.listSort.length; i++) {
      cnt += moveObjectsToBack(document, di, storage, this.listSort[i]);
    }
    di.repaintViews();
    EAction.handleUserMessage(qsTr("object ordering done, objects ordered: ") + cnt);

    //--cleanup
    dialog.destroy();
    this.terminate();
};

ReorderObjects.prototype.rebuildLists = function() {
    this.widgets["listAvail"].clear();
    this.widgets["listSort"].clear();
    for (var i = 0; i < this.listAvail.length; i++) {
      this.widgets["listAvail"].addItem(this.listAvail[i]);
    }
    for (var i = 0; i < this.listSort.length; i++) {
      this.widgets["listSort"].addItem(this.listSort[i]);
    }
    this.checkButtonEnable();
}

ReorderObjects.prototype.checkButtonEnable = function() {
    var sortSel = this.widgets["listSort"].selectedItems().length > 0;
    this.widgets["goTop"].enabled = sortSel;
    this.widgets["goUp"].enabled = sortSel;
    this.widgets["goDown"].enabled = sortSel;
    this.widgets["goBottom"].enabled = sortSel;
    this.widgets["invertOrder"].enabled = this.listSort.length > 0;
    this.widgets["noSort"].enabled = sortSel;
    this.widgets["doSort"].enabled = this.widgets["listAvail"].selectedItems().length > 0;
    this.widgets["buttonBox"].button(QDialogButtonBox.Ok).enabled = this.listSort.length > 0;
};

ReorderObjects.prototype.goTop = function() {
    var move = this.widgets["listSort"].selectedItems();
    if (move.length == 0 || this.widgets["listSort"].row(move[0]) == 0) {
      return;
    }
    for (var i = 0; i < move.length; i++) {
      var ind = this.widgets["listSort"].row(move[i]);
      this.listSort.splice(ind, 1);
      this.listSort.splice(i, 0, move[i].text());
      var item = this.widgets["listSort"].takeItem(ind);
      this.widgets["listSort"].insertItem(i, move[i]);
      move[i].setSelected(true);
    }
};

ReorderObjects.prototype.goUp = function() {
    var move = this.widgets["listSort"].selectedItems();
    if (move.length == 0 || this.widgets["listSort"].row(move[0]) == 0) {
      return;
    }
    for (var i = 0; i < move.length; i++) {
      var ind = this.widgets["listSort"].row(move[i]);
      this.listSort.splice(ind, 1);
      this.listSort.splice(ind - 1, 0, move[i].text());
      var item = this.widgets["listSort"].takeItem(ind);
      this.widgets["listSort"].insertItem(ind - 1, move[i]);
      move[i].setSelected(true);
    }
};

ReorderObjects.prototype.goDown = function() {
    var move = this.widgets["listSort"].selectedItems();
    if (move.length == 0 || this.widgets["listSort"].row(move[move.length - 1]) == this.listSort.length - 1) {
      return;
    }
    for (var i = 0; i < move.length; i++) {
      var ind = this.widgets["listSort"].row(move[i]);
      this.listSort.splice(ind, 1);
      this.listSort.splice(ind + move.length, 0, move[i].text());
      var item = this.widgets["listSort"].takeItem(ind);
      this.widgets["listSort"].insertItem(ind + move.length, move[i]);
      move[i].setSelected(true);
    }
};

ReorderObjects.prototype.goBottom = function() {
    var move = this.widgets["listSort"].selectedItems();
    if (move.length == 0 || this.widgets["listSort"].row(move[move.length - 1]) == this.listSort.length - 1) {
      return;
    }
    for (var i = 0; i < move.length; i++) {
      var ind = this.widgets["listSort"].row(move[i]);
      this.listSort.splice(ind, 1);
      this.listSort.splice(this.listSort.length, 0, move[i].text());
      var item = this.widgets["listSort"].takeItem(ind);
      this.widgets["listSort"].insertItem(this.widgets["listSort"].count, move[i]);
      move[i].setSelected(true);
    }
};

ReorderObjects.prototype.invertOrder = function() {
    var newList = new Array();
    for (var i = this.listSort.length - 1; i >= 0; i--) {
      newList.push(this.listSort[i]);
    }
    this.listSort = newList;
    this.rebuildLists();
};

ReorderObjects.prototype.noSort = function() {
    var move = this.widgets["listSort"].selectedItems();
    for (var i = move.length - 1; i >= 0; i--) {
      removeFromStringArray(this.listSort, move[i].text());
      this.listAvail.push(move[i].text());
    }
    this.rebuildLists();
};

ReorderObjects.prototype.doSort = function() {
    var move = this.widgets["listAvail"].selectedItems();
    for (var i = move.length - 1; i >= 0; i--) {
      removeFromStringArray(this.listAvail, move[i].text());
      this.listSort.push(move[i].text());
    }
    this.rebuildLists();
};

function removeFromStringArray(arr, what) {
    var found = arr.indexOf(what);
    while (found !== -1) {
      arr.splice(found, 1);
      found = arr.indexOf(what);
    }
}

ReorderObjects.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.ReorderObjects, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/ReorderObjects.js");
    action.setIcon(basePath + "/ReorderObjects.svg");
    action.setDefaultShortcut(new QKeySequence("q,j"));
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupModification);
    action.setSortOrder(QCTrack.MenuOrder.ReorderObjects);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

