/*
 * This script selects all tracks "connected" to the currently
 * selected tracks, ie. have low distance at their ends.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-09-02
 */

include("../QCTrack.js");

function SelectCon(guiAction) {
    EAction.call(this, guiAction);
}

SelectCon.prototype = new EAction();
SelectCon.includeBasePath = includeBasePath;

SelectCon.prototype.beginEvent = function() {
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();

    var dialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/SelectCon",
                   "SelectConDialog.ui");
    WidgetFactory.restoreState(dialog);
    var widgets = getWidgets(dialog);
    widgets["buttonBox"].button(QDialogButtonBox.RestoreDefaults)["clicked()"].connect(this, function() { WidgetFactory.restoreState(dialog, null, null, true); });
    widgets["buttonBox"].button(QDialogButtonBox.Ok).setFocus();
    if (!dialog.exec()) { //the user aborted
        dialog.destroy();
        this.terminate();
        return;
    }
    WidgetFactory.saveState(dialog);
    this.sameLayer = widgets["stayOnLayer"].checked;
    this.stopAtTurnout = widgets["stopAtTurnout"].checked;
    this.maxDist = widgets["distanceRestrict"].getValue();
    dialog.destroy();
    if (this.maxDist < 0) {
      EAction.handleUserMessage(qsTr("distance must not be negative"));
      (new QMessageBox(null, qsTr("Select Connected Track"), qsTr("distance must not be negative"),
        new QMessageBox.StandardButtons(QMessageBox.Ok), null, Qt.Dialog)).exec();
      this.terminate();
      return;
    }

    this.addSelectionIds = new Array();
    var allIds = document.queryAllBlockReferences();
    var allTracks = new Array(); //entries are three-field-arrays containing a block, an array with all TR_END-lines inside that block, and the selection state
    for (var i = 0; i < allIds.length; i++) { //first run: copy all blocks with TR_END lines
      var id = allIds[i];
      var block = document.queryEntity(id);
      var data = block.getData();
      var subIds = document.queryBlockEntities(data.getReferencedBlockId());
      var lines = new Array();
      for (var j = 0; j < subIds.length; j++) {
        var line = data.queryEntity(subIds[j]);
        if (isLineEntity(line) && line.getLayerName() == "TR_ENDS") {
          lines.push(line);
        }
      }
      if (lines.length > 0) {
        allTracks.push([block, lines, block.isSelected()]);
      }
    }
    for (var i = 0; i < allTracks.length; i++) { //second run: recursively check selected blocks
      var block = allTracks[i];
      if (block[0].isSelected()) {
        this.recursiveSelect(allTracks, i);
      }
    }
    di.selectEntities(this.addSelectionIds, true);
    EAction.handleUserMessage(qsTr("added ") + this.addSelectionIds.length + qsTr(" tracks to selection."));

    this.terminate();
};

SelectCon.prototype.recursiveSelect = function(allTracks, cmp) {
    for (var i = 0; i < allTracks.length; i++) {
      if (i == cmp || allTracks[i][2] || (this.stopAtTurnout && allTracks[i][1].length>2)) {
        continue;
      }
      this.selectIfConnected(allTracks, cmp, i);
    }
};

SelectCon.prototype.selectIfConnected = function(allTracks, cmp, to) {
  if (this.sameLayer && allTracks[cmp][0].getLayerName() != allTracks[to][0].getLayerName()) {
    return;
  }
  var cmpLines = allTracks[cmp][1];
  var toLines = allTracks[to][1];
  for (var i = 0; i < cmpLines.length; i++) {
    var cmpLine = cmpLines[i];
    for (var j = 0; j < toLines.length; j++) {
      var line = toLines[j];
      if (line.getStartPoint().getDistanceTo(cmpLine.getEndPoint()) <= this.maxDist
          && line.getEndPoint().getDistanceTo(cmpLine.getStartPoint()) <= this.maxDist) {
        this.addSelectionIds.push(allTracks[to][0].getId());
        allTracks[to][2] = true;
        this.recursiveSelect(allTracks, to);
        return;
      }
    }
  }
};

SelectCon.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.SelectCon, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setRequiresSelection(true);
    action.setScriptFile(basePath + "/SelectCon.js");
    action.setIcon(basePath + "/SelectCon.svg");
    action.setDefaultShortcut(new QKeySequence("q,s"));
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupModification);
    action.setSortOrder(QCTrack.MenuOrder.SelectCon);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
    var appWin = EAction.getMainWindow();
    appWin.addSelectionListener(action);
};

