/*
 * This script variegates track combinations with parameters
 * according to the user dialog.
 * Attention! This script is highly re-used by FillStraight, don't
 * change it without checking FillStraight!
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-04-20
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");
include("../../WidgetFactory.js");

function Variegate(guiAction) {
    QCTrack.call(this, guiAction);
}

Variegate.prototype = new EAction();
Variegate.includeBasePath = includeBasePath;

Variegate.State = {
    SettingPosition : 0
};

Variegate.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);

    if (!this.doVariegation(undefined)) {
      this.terminate();
      return;
    }

    //give control back to the user
    EAction.handleUserMessage("select position where to place " + this.created + " combinations");
    this.setState(Variegate.State.SettingPosition);
    //do not terminate, user has to set the target point
};

Variegate.prototype.doVariegation = function(destinationLine) {
    var document = EAction.getDocument();

    //get a list of all (straight) tracks
    this.list = new Array(); //each element contains a five-field array containing 0: the block name, 1: the block id, 2: max. usage, 3: its length, and 4: the amount
    var allIds = document.queryAllBlocks();
    var canGenerate = true;
    this.objDistBase = 0;
    for (var i = 0; i < allIds.length; i++) { //first run: collect blocks with exactly one straight line in TR_MIDS
      var id = allIds[i];
      var block = document.queryBlockDirect(id);
      var blockName = block.getName();
      if (blockName[0] == '*') { //reserved blocks like paper space
        continue;
      }
      var k = 0;
      while (k < this.list.length) {
        if (this.list[k][0] == blockName) { //block already existing
          break;
        }
        k++;
      }
      if (k == this.list.length) {
        var subIds;
        subIds = document.queryBlockEntities(id);
        var blockLen = 0;
        var blockHeight = 0;
        var blockEnds = 0;
        var thisCanNotGenerate = false;
        for (var j = 0; j < subIds.length; j++) {
          var sub = document.queryEntity(subIds[j]);
          if (sub.getLayerName() == "TR_MIDS") {
            if (blockLen > 0) { //block has another entity on TR_MIDS layer
              blockLen = 0;
              //EAction.handleUserMessage("block " + blockName + " contains more than one mid element, ignoring it");
              break;
            }
            if (isLineEntity(sub)) {
              blockLen = sub.getStartPoint().getDistanceTo(sub.getEndPoint());
              if (sub.getAngle() != 0) { //can only generate horizontal tracks
                //EAction.handleUserMessage("line in block " + blockName + " has angle != 0, can only print (will not generate tracks)");
                thisCanNotGenerate = true;
              }
            }
            else { //not (just) a straight line
              break;
            }
          }
          else {
            var points = sub.getEndPoints();
            for (var k = 0; k < points.length; k++) {
              var ay = Math.abs(points[k].getY());
              if (ay > blockHeight) {
                blockHeight = ay;
              }
            }
            if (sub.getLayerName() == "TR_ENDS") {
              blockEnds++;
            }
          }
        }
        if (blockLen > 0 && blockEnds == 2) {
          if (thisCanNotGenerate) {
            canGenerate = false;
          }
          this.list.push([blockName, id, 0, blockLen, 0]);
          if (blockHeight > this.objDistBase) {
            this.objDistBase = blockHeight;
          }
        }
      }
    }
    if (this.list.length == 0) {
      EAction.handleUserMessage(qsTr("no straight line for variegation available"));
      (new QMessageBox(null, qsTr("Variegate"), qsTr("no straight line for variegation available"),
        new QMessageBox.StandardButtons(QMessageBox.Ok), null, Qt.Dialog)).exec();
      return false;
    }
    this.list.sort(function(a, b) {
      return a[0].localeCompare(b[0]);
    });

    //let the user enter the desired parameters
    var paramdialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/Variegate",
                   "VariegateDialog.ui");
    var widgets = getWidgets(paramdialog);
    var dialist = widgets["list"];
    var maxElemSpin;
    {
      var w = new QWidget();
      var vbl = new QHBoxLayout(w);
      var item;
      vbl.addWidget(new QLabel(qsTr("overall elements")), 0, 0);
      vbl.addWidget(item = new QLabel("max"), 0, 0); 
      vbl.addWidget(item = new QSpinBox(), 0, 0);
      item.minimum = -1;
      item.maximum = 999;
      item.value = 16;
      item.alignment = Qt.AlignVCenter | Qt.AlignRight;
      maxElemSpin = item;
      w.setLayout(vbl);
      var lwi = new QListWidgetItem(dialist);
      lwi.setSizeHint(w.sizeHint);
      dialist.addItem(lwi);
      dialist.setItemWidget(lwi, w);
    }
    var maxSpins = new Array();
    for (var i = 0; i < this.list.length; i++) {
      var w = new QWidget();
      var vbl = new QHBoxLayout(w);
      var item;
      vbl.addWidget(new QLabel(this.list[i][0]), 0, 0);
      vbl.addWidget(item = new QLabel(this.list[i][3].toFixed(1)), 0, 0); 
      item.alignment = Qt.AlignVCenter | Qt.AlignRight;
      vbl.addWidget(item = new QSpinBox(), 0, 0);
      item.minimum = -1;
      item.maximum = 999;
      item.value = 0;
      item.alignment = Qt.AlignVCenter | Qt.AlignRight;
      maxSpins.push(item);
      w.setLayout(vbl);
      var lwi = new QListWidgetItem(dialist);
      lwi.setSizeHint(w.sizeHint);
      dialist.addItem(lwi);
      dialist.setItemWidget(lwi, w);
    }
    var selected = document.querySelectedEntities();
    if (destinationLine == undefined && selected.length > 0) {
      for (var i = 0; i < selected.length; i++) {
        var blockName = undefined;
        var id = selected[i];
        var block = document.queryEntity(id);
        if (!isBlockReferenceEntity(block)) {
          continue;
        }
        blockName = block.getReferencedBlockName();
        if (blockName == undefined) {
          continue;
        }
        for (var j = 0; j < this.list.length; j++) {
          if (this.list[j][0] == blockName) {
            maxSpins[j].value = -1;
          }
        }
      }
    }
    else {
      for (var i = 0; i < this.list.length; i++) {
        maxSpins[i].value = -1;
      }
    }
    widgets["useNone"]["clicked()"].connect(this, function() {
      for (var i = 0; i < this.list.length; i++) {
        maxSpins[i].value = 0;
      }
    });
    widgets["useAll"]["clicked()"].connect(this, function() {
      for (var i = 0; i < this.list.length; i++) {
        maxSpins[i].value = -1;
      }
    });
    if (destinationLine != undefined) {
      var destLen = destinationLine.getLength();
      widgets["minLen"].setText((destLen > 5 ? destLen - 5 : 0).toFixed(2));
      widgets["desiredLen"].setText(destLen.toFixed(2));
      widgets["maxLen"].setText((destLen + 5).toFixed(2));
    }
    do {
      if (!paramdialog.exec()) { //the user aborted
        paramdialog.destroy();
        return false;
      }
      this.maxElems = maxElemSpin.value;
      for (var i = this.list.length - 1; i >= 0; i--) {
        if (maxSpins[i].value == 0) {
          this.list.splice(i, 1);
          maxSpins.splice(i, 1);
        }
        else {
          this.list[i][2] = maxSpins[i].value;
        }
      }
      if (this.list.length == 0) { //the user deselected all tracks
        EAction.handleUserMessage(qsTr("all tracks were disabled by user, can not variegate"));
        (new QMessageBox(null, qsTr("Variegate"), qsTr("all tracks were disabled by user, can not variegate"),
          new QMessageBox.StandardButtons(QMessageBox.Ok), null, Qt.Dialog)).exec();
        paramdialog.destroy();
        return false;
      }
      this.minLen = widgets["minLen"].getValue();
      this.desiredLen = widgets["desiredLen"].getValue();
      this.maxLen = widgets["maxLen"].getValue();

      //produce combinations with backtracking
      EAction.handleUserMessage("variegating with " + this.list.length + " tracks...");
      this.solutions = new Array; //will contain all arrays with [0]: array containing length and element track and [1]: solution arrays particular counts
      this.extendCurrentCombination(0); //do all the work
      this.solutions.sort(function(a, b) {
        var lenDiff = a[0][0] - b[0][0];
        if (lenDiff != 0) {
          return lenDiff;
        }
        return a[0][1] - b[0][1]; //if length is equal, compare element count
      });

      if (this.solutions.length == 0) {
        if ((new QMessageBox(null, qsTr("Variegate"), qsTr("sorry, did not found a solution - retry?"),
            new QMessageBox.StandardButtons(QMessageBox.Yes, QMessageBox.No), null, Qt.Dialog)).exec() != QMessageBox.Yes) {
          paramdialog.destroy();
          this.terminate();
          return;
        }
      }
    } while (this.solutions.length == 0);

    //show results and get further action
    var resultdialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/Variegate",
                   "VariegateResultDialog.ui");
    var widgets = getWidgets(resultdialog);
    this.okButton = widgets["buttonBox"].button(QDialogButtonBox.Ok);
    widgets["resultLabel"].text = qsTr("Select solutions, searched: ") + this.desiredLen.toFixed(1);
    this.filtable = widgets["filterTable"];
    this.filtable.clear();
    this.filtable.maximumHeight = 110;
    this.filtable.columnCount = this.list.length + 1;
    this.filtable.rowCount = 2;
    this.filtable.setVerticalHeaderItem(0, new QTableWidgetItem("max"));
    this.filtable.setVerticalHeaderItem(1, new QTableWidgetItem("min"));
    this.multiFilterReset = true;
    this.filMinSpins = new Array();
    this.filMaxSpins = new Array();
    for (var i = 0; i < this.list.length; i++) {
      this.filtable.setHorizontalHeaderItem(i, new QTableWidgetItem(this.list[i][0]));
      var minCnt = this.solutions[0][1][i];
      var maxCnt = this.solutions[0][1][i];
      for (var si = 1; si < this.solutions.length; si++) {
        if (this.solutions[si][1][i] < minCnt) {
          minCnt = this.solutions[si][1][i];
        }
        if (this.solutions[si][1][i] > maxCnt) {
          maxCnt = this.solutions[si][1][i];
        }
      }
      this.filtable.setCellWidget(0, i, item = this.createFilterSpinBox(minCnt, maxCnt, maxCnt));
      this.filMaxSpins.push(item);
      this.filtable.setCellWidget(1, i, item = this.createFilterSpinBox(minCnt, maxCnt, minCnt));
      this.filMinSpins.push(item);
    }
    var minSum = this.solutions[0][0][1];
    var maxSum = this.solutions[0][0][1];
    for (var i = 1; i < this.solutions.length; i++) {
      if (this.solutions[i][0][1] < minSum) {
        minSum = this.solutions[i][0][1];
      }
      if (this.solutions[i][0][1] > maxSum) {
        maxSum = this.solutions[i][0][1];
      }
    }
    this.filtable.setHorizontalHeaderItem(this.list.length, new QTableWidgetItem("sum"));
    this.filtable.setCellWidget(0, this.list.length, item = this.createFilterSpinBox(minSum, maxSum, maxSum));
    this.maxElemSpin = item;
    this.filtable.setCellWidget(1, this.list.length, item = this.createFilterSpinBox(minSum, maxSum, minSum));
    this.minElemSpin = item;
    this.multiFilterReset = false;
    this.restable = widgets["resultTable"];
    this.rebuildResultTable();
    widgets["showAll"]["clicked()"].connect(this, function() {
      this.multiFilterReset = true;
      for (var i = 0; i < this.list.length; i++) {
        this.filMinSpins[i].value = this.filMinSpins[i].minimum;
        this.filMaxSpins[i].value = this.filMaxSpins[i].maximum;
      }
      this.minElemSpin.value = this.minElemSpin.minimum;
      this.maxElemSpin.value = this.maxElemSpin.maximum;
      this.multiFilterReset = false;
      this.rebuildResultTable();
    });
    widgets["exportSelected"]["clicked()"].connect(this, function() {
      this.solutionToText(false, false);
    });
    widgets["exportAll"]["clicked()"].connect(this, function() {
      this.solutionToText(false, true);
    });
    widgets["printSelected"]["clicked()"].connect(this, function() {
      this.solutionToText(true, false);
    });
    widgets["printAll"]["clicked()"].connect(this, function() {
      this.solutionToText(true, true);
    });
    widgets["orthoDist"].setText((this.objDistBase * 4).toFixed(2));
    if (destinationLine != undefined) {
      widgets["resultTable"].selectionMode = QAbstractItemView.SingleSelection;
      widgets["orthoDistLabel"].enabled = false;
      widgets["orthoDist"].enabled = false;
      widgets["labelTracks"].enabled = false;
    }
    if (!resultdialog.exec()) { //the user aborted
      paramdialog.destroy();
      resultdialog.destroy();
      return false;
    }

    //ok, the user wants us to build tracks in QCad, create the object arrays
    this.createLabels = new Array(); //will contain all label-information with [0]: x, [1]: y, [2]: label text
    this.createBlocks = new Array(); //will contain all block-information with [0]; x, [1]; y, [2]: block id
    this.created = 0;
    var rows = this.restable.selectionModel().selectedRows();
    var distance = widgets["orthoDist"].getValue();
    var doLabel = widgets["labelTracks"].checked;
    for (var i = 0; i < this.solutions.length; i++) {
      if (rows == undefined || this.isInRowSelection(rows, i)) {
        var y = -distance * this.created;
        var solPck = this.solutions[i];
        if (doLabel) {
          this.createLabels.push([-distance / 4, y, solPck[0][0].toFixed(1)]);
        }
        this.pushBlocksOfSolution(0, y, solPck[1]);
        this.created++;
      }
    }
    paramdialog.destroy();
    resultdialog.destroy();
    return true;
};

Variegate.prototype.createFilterSpinBox = function(imin, imax, ival) {
    var item = new QSpinBox()
    item.minimum = imin;
    item.maximum = imax;
    item.value = ival;
    item.alignment = Qt.AlignVCenter | Qt.AlignRight;
    item["valueChanged(int)"].connect(this, function(val) {
      if (!this.multiFilterReset) {
        this.rebuildResultTable();
      }
    });
    return item;
};

Variegate.prototype.rebuildResultTable = function() {
    this.shownResults = new Array();

    var bestInd = 0;
    for (var i = 0; i < this.solutions.length; i++) {
      if (this.solutions[i][0][1] >= this.minElemSpin.value && this.solutions[i][0][1] <= this.maxElemSpin.value) {
        var isValid = true;
        for (var ti = 0; ti < this.list.length; ti++) {
          if (this.solutions[i][1][ti] < this.filMinSpins[ti].value || this.solutions[i][1][ti] > this.filMaxSpins[ti].value) {
            isValid = false;
            break;
          }
        }
        if (isValid) {
          this.shownResults.push(i);
          var lenDiff = Math.abs(this.desiredLen - this.solutions[i][0][0]) - Math.abs(this.desiredLen - this.solutions[bestInd][0][0]); //compare length
          if (lenDiff < 0 || (lenDiff == 0 && this.solutions[i][0][1] < this.solutions[bestInd][0][1])) { //if length is equal, compare element count
            bestInd = i;
          }
        }
      }
    }
    this.restable.clear();
    this.restable.columnCount = this.list.length + 1;
    this.restable.rowCount = this.shownResults.length;
    for (var i = 0; i < this.list.length; i++) {
      this.restable.setHorizontalHeaderItem(i, new QTableWidgetItem(this.list[i][0]));
    }
    this.restable.setHorizontalHeaderItem(this.list.length, new QTableWidgetItem("sum"));
    var toSelect = 0;
    for (var i = 0; i < this.shownResults.length; i++) {
      var solInd = this.shownResults[i];
      if (solInd == bestInd) {
        toSelect = i;
      }
      var solution = this.solutions[solInd];
      var item;
      var diff = solution[0][0] - this.desiredLen;
      this.restable.setVerticalHeaderItem(i, item = new QTableWidgetItem(solution[0][0].toFixed(1)
        + " (" + (diff < 0 ? diff.toFixed(1) : "+" + diff.toFixed(1)) + ")" + " / " + solution[0][1]));
      item.setTextAlignment(Qt.AlignVCenter | Qt.AlignRight);
      this.restable.setItem(i, this.list.length, item = new QTableWidgetItem(solution[0][1].toFixed(0)));
      item.setTextAlignment(Qt.AlignVCenter | Qt.AlignRight);
      for (var ti = 0; ti < this.list.length; ti++) {
        this.restable.setItem(i, ti, item = new QTableWidgetItem(solution[1][ti].toFixed(0)));
        item.setTextAlignment(Qt.AlignVCenter | Qt.AlignRight);
      }
    }
    if (this.shownResults.length == 0) {
      this.okButton.enabled = false;
    }
    else {
      this.okButton.enabled = true;
      this.restable.selectRow(toSelect);
    }
};

Variegate.prototype.solutionToText = function(onlyPrint, all) {
    var rows = undefined;
    if (!all) {
      rows = this.restable.selectionModel().selectedRows();
    }
    if (onlyPrint) { //directly print to QCad console
      for (var i = 0; i < this.solutions.length; i++) {
        if (rows == undefined || this.isInRowSelection(rows, i)) {
          var solPck = this.solutions[i];
          EAction.handleUserMessage("one solution has " + solPck[0][1]
            + " elements coming to length " + solPck[0][0].toFixed(1) + ":");
          this.printSolution(solPck[1]);
        }
      }
    }
    else { //export to clipboard
      var result = "len,elem";
      for (var i = 0; i < this.list.length; i++) {
        result += "," + this.list[i][0];
      }
      result += "\n";
      for (var i = 0; i < this.solutions.length; i++) {
        if (rows == undefined || this.isInRowSelection(rows, i)) {
          var solPck = this.solutions[i];
          result += solPck[0][0].toFixed(1) + "," + solPck[0][1];
          for (var j = 0; j < this.list.length; j++) {
            result += "," + solPck[1][j];
          }
          result += "\n";
        }
      }
      getClipboard().setText(result);
    }
};

Variegate.prototype.isInRowSelection = function(rows, index) {
    for (var i = 0; i < rows.length; i++) {
      if (this.shownResults[rows[i].row()] == index) {
        return true;
      }
    }
    return false;
};

Variegate.prototype.extendCurrentCombination = function(startAt) {
    var curLen = this.getCurrentLength();
    var curElems = this.getCurrentElementCount();
    if (curLen > this.maxLen || (this.maxElems > -1 && curElems > this.maxElems)) { //exceeded maximum length or maximum element count
      return;
    }
    if (startAt < this.list.length) {
      while ((this.list[startAt][2] == -1 || this.list[startAt][4] <= this.list[startAt][2]) && curLen <= this.maxLen) { //maximum element count and maximum length must not be exceeded
        this.extendCurrentCombination(startAt + 1);
        this.list[startAt][4]++;
        curLen = this.getCurrentLength();
        curElems = this.getCurrentElementCount();
      }
      this.list[startAt][4]=0; //reset for further back-tracking
    }
    else { //all elements done (maybe some are ==0), check if solution found
      if (curLen >= this.minLen) { //found a solution
        var sol = this.getSolutionArray();
        var len = this.getCombinationLength(sol);
        var elem = this.getCombinationElementCount(sol);
        this.solutions.push([[len, elem], sol]);
      }
    }
};

Variegate.prototype.getCurrentLength = function() {
    var len = 0;
    for (var i = 0; i < this.list.length; i++) {
      len += this.list[i][3] * this.list[i][4];
    }
    return len;
};

Variegate.prototype.getCurrentElementCount = function() {
    var elems = 0;
    for (var i = 0; i < this.list.length; i++) {
      elems += this.list[i][4];
    }
    return elems;
};

Variegate.prototype.getCombinationLength = function(sol) {
    var len = 0;
    for (var i = 0; i < this.list.length; i++) {
      len += this.list[i][3] * sol[i];
    }
    return len;
};

Variegate.prototype.getCombinationElementCount = function(sol) {
    var elems = 0;
    for (var i = 0; i < this.list.length; i++) {
      elems += sol[i];
    }
    return elems;
};

Variegate.prototype.printSolution = function(sol) {
    for (var i = 0; i < this.list.length; i++) {
      EAction.handleUserMessage("-- " + this.list[i][0] + ": " + this.list[i][3] + " * " + sol[i] + " => " + (this.list[i][3] * sol[i]).toFixed(1));
    }
};

Variegate.prototype.pushBlocksOfSolution = function(x, y, sol) {
    for (var i = 0; i < this.list.length; i++) {
      var cnt = sol[i];
      var blockId = this.list[i][1];
      var len = this.list[i][3];
      while (cnt > 0) {
        this.createBlocks.push([x, y, blockId]);
        x += len;
        cnt--;
      }
    }
};

Variegate.prototype.getSolutionArray = function() {
    var sol = new Array(this.list.length);
    for (var i = 0; i < this.list.length; i++) {
      sol[i] = this.list[i][4];
    }
    return sol;
};

//--- create and place tracks

Variegate.prototype.setState = function(state) {
    EAction.prototype.setState.call(this, state);

    this.setCrosshairCursor();
    this.getDocumentInterface().setClickMode(RAction.PickCoordinate);

    var appWin = RMainWindowQt.getMainWindow();
    var trPos = qsTr("Position");
    this.setCommandPrompt(trPos);
    this.setLeftMouseTip(trPos);
    this.setRightMouseTip(EAction.trCancel);
    EAction.showSnapTools();
};

Variegate.prototype.pickCoordinate = function(event, preview) {
    var di = this.getDocumentInterface();
    var pos = event.getModelPosition();

    this.targetPosition = pos;
    this.targetAngle = 0;

    if (preview) {
      di.previewOperation(this.getOperation(true));
    }
    else {
      var op = this.getOperation(false);
      if (!isNull(op)) {
        di.applyOperation(op);
        di.setRelativeZero(pos);
        this.setState(-1);
        this.terminate();
      }
    }
};

Variegate.prototype.addBlockOperations = function(document, op, preview) {
    var nextFreeId = document.getStorage().getMaxObjectId();
    for (var i = 0; i < this.createBlocks.length; i++) { //create all blocks
      var coa = this.createBlocks[i];
      if (preview && coa[1] != 0) break; //preview only for first line
      var blockId = coa[2];
      var blockReferenceData = new RBlockReferenceData();
      blockReferenceData.setReferencedBlockId(blockId);
      blockReferenceData.setPosition(new RVector(coa[0], coa[1]));
      blockReferenceData.setScaleFactors(new RVector(1.0, 1.0));
      blockReferenceData.setRotation(0.0);
      blockReferenceData.move(this.targetPosition);
      if (this.angle != 0) {
        blockReferenceData.rotate(this.targetAngle, this.targetPosition);
      }
      var blockRef = new RBlockReferenceEntity(document, blockReferenceData);
      op.addObject(blockRef);
    }
};

Variegate.prototype.getOperation = function(preview) {
    var document = this.getDocument();
    var op = new RAddObjectsOperation();

    this.addBlockOperations(document, op, preview);

    for (var i = 0; i < this.createLabels.length; i++) { //create all labels
      var cla = this.createLabels[i];
      if (preview && i > 2) break; //preview only for first three line
      var pos = new RVector(this.targetPosition.getX() + cla[0], this.targetPosition.getY() + cla[1]);
      var text = new RTextEntity(document, new RTextData(
        pos,                 // position
        pos,                 // alignment point
        this.objDistBase,    // height
        2.0,                 // text width (ignored for now)
        RS.VAlignMiddle,     // alignments
        RS.HAlignRight,
        RS.LeftToRight,
        RS.Exact,
        1.0,                 // line spacing factor
        cla[2],       // the text
        "Arial",      // font
        false,        // bold
        false,        // italic
        0.0,          // angle
        false         // simple text without formatting
      ));
      op.addObject(text);
    }
    return op;
};

//--- init

Variegate.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.Variegate, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/Variegate.js");
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupOther);
    action.setSortOrder(QCTrack.MenuOrder.Variegate);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

